// Aner Ben-Artzi - Efficient Coherence-Based Visibility Sampling (CBVS) of Environment Maps
// Columbia University 2004
//
// Area.h: interface for the Area class.
//////////////////////////////////////////////////////////////////////

/*
The screen is divided into 6 areas.  Each area can contain a Trace or Closeup
*/

#if !defined(AFX_AREA_H__59E5DE0D_5CC9_427E_AEF1_499FED48B6C9__INCLUDED_)
#define AFX_AREA_H__59E5DE0D_5CC9_427E_AEF1_499FED48B6C9__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "VVTrace.h"

//(x,y) = (0,0) is the bottom-left

#define I(row,col) (((size-1)-(row))*size+(col))

class Area  
{
public:
	//if true then passive mouse motion doesn't affect this area
	bool lockFocus;
	//if the type is ONELIGHT, this is the light off which it was based
	int lightused;
	//if this is a closeup, indicates which corner it shows
	int corner;
	//indicates if this area is an image of the lightspace
	bool closeup;
	//this is the area off which this area was based, if any.
	Area* parentArea;
	//the center pixel that is currently in focus
	//if mousetrack then changes with passive motion
	//if not mousetracking than changes with arrow keyboard press
	int activex;
	int activey;
	//top left location of the text.
	//changes when mousetrack is off an mouse is clicked
	int textx;
	int texty;
	//list of pixels that are highlighted
	Point3d* highlighted[40];
	//indicates if highlighting is brighten or make-yellow
	bool highlightyellow;
	//number of currently highlighted pixels
	int numhighlighted;
	//when the area is a closeup, indicates the active light
	int activeLight;
	//indicates the type of area
	AreaType type;
	//if the area is a trace, points to the Trace object
	VVTrace* trace;
	//indicates if the area values should be normalized
	bool normalize;
	//the actual pixels of the area
	Point3d* area;
	//the size of the area
	int size;

public:
	void brighten(float amount);
	void setType(AreaType thetype);
	void setParent(Area* parea);
	void setFocus(int x, int y);
	void unFocus();
	Area(int thesize);
	virtual ~Area();
	void fillArea();

};

#endif // !defined(AFX_AREA_H__59E5DE0D_5CC9_427E_AEF1_499FED48B6C9__INCLUDED_)
