// Aner Ben-Artzi - Efficient Coherence-Based Visibility Sampling (CBVS) of Environment Maps
// Columbia University 2004
//// Container.cpp: implementation of the Container class.


#include "Container.h"
#include "VVTrace.h"

#ifndef RAYTRACER
#include "Area.h"
#endif //RAYTRACER

////////////////////////////////////////////////////////////////////////
// Static variables
///////////////////////////////////////////////////////////////////////
bool Container::useGlut = true;
int Container::oldx;
int Container::oldy;
bool Container::keys[255];
int Container::keysdown;
bool Container::record = false;
FILE* Container::recordFile = NULL;
char Container::previousRecord[40];
int Container::numberOfLights;
int Container::size;
Point3d Container::cameraLocation;
Point3d Container::cameraLookAt;
Point3d Container::objectColor[20];
bool Container::trackMouse = true;
int Container::strongestSources[20];

#ifndef RAYTRACER
VVTrace* Container::theOneTrueTrace;
Area* Container::areas[6] = {NULL, NULL, NULL, NULL, NULL, NULL};
#endif //RAYTRACER

Depth_Normal** Image_Info;
int Container::activeArea = 0;
Light_Graph** graph;
Light* thelights;
bool* Container::horizons;
int Container::draggedArea;
int** Container::lightID;
int Container::lightSize;
int Container::glutHeight;
int Container::glutWidth;
bool Container::leftDown = false;
bool Container::passiveParent = false;
bool Container::dragAllLights = false;
int Container::draggedx = 0;
int Container::draggedy = 0;


///////////////////////////////////////////////////////////////////////////
// Member functions
//////////////////////////////////////////////////////////////////////////
Container::Container()
{

}

Container::~Container()
{
#ifndef RAYTRACER
	delete theOneTrueTrace;
#endif //RAYTRACER
	for (int i = 0; i < lightSize; i++) {
		delete[] lightID[i];
	}
	delete[] lightID;
	delete[] horizons;
}

int Container::findNearestLight(Point3d direction, bool scoot) {
//	Point3d proj;
	Point3d localdir;
	P3D_Assign_Vector(localdir, direction);
	if (localdir[1] <= 0) { //becuase we only have the upper hemisphere
		if (!scoot) {
			return -1;
		}

		localdir[1] = 0;
		P3D_VNormalizeEq(localdir);

		while (P3D_VSumSqr(localdir) > 1) {
			P3D_VScaleEq(localdir, 1-P3D_EPSILON);
		}	
	}
//	projectStretch(proj, localdir);
//	int j = (int)((((-1*proj[0])+1)/2) * (lightSize-1));
//	int i = (int)(((proj[1]+1)/2) * (lightSize-1));
	
	int i, j;

	direction2xy(i, j, localdir, lightSize);

	if (j >= 0 && i >= 0 && j < lightSize && i < lightSize && lightID[j][i] >= 0) {
			return lightID[j][i];
	} else {
		for (int b = -1; b < 2; b++) {
			for (int c = -1; c < 2; c++) {
				if (j+b >= 0 && i+c >= 0 && j+b < lightSize && i+c < lightSize) {
					if (lightID[j+b][i+c] >= 0) {
						return lightID[j+b][i+c];
					}
				}
			}
		}
	}
	printf("Error: findNearestLight");
	return -2;
};


void Container::init() {
	printf("Initializing\n");
	for (int f = 0; f < NUMFLAGS; f++) {
		VVTrace::globalFlags[f] = false;
	}

	//get the right light resolution:
	int h;

	for (h = 0; h < numberOfLights; h++) {
		P3D_VNormalizeEq(thelights[h].direction);
	}

	printf("Finding nearest light neighbors\n");
	for (h = 0; h < numberOfLights; h++) {
		printf("Light %d of %d   %c", h, numberOfLights, h == numberOfLights-1 ? '\n' : '\r');
		float mindist = -1;
		int minLight = -2;
		float dist;
		for (int hInner = 0; hInner < numberOfLights; hInner++) {
			if (h == hInner) {
				continue;
			}
			P3D_VDot(dist, thelights[h].direction, thelights[hInner].direction);
			if (dist > mindist) {
				minLight = hInner;
				mindist = dist;
			}
		}
		thelights[h].nearestNeighborNum = minLight;
		thelights[h].nearestNeighborDist = acos(mindist); //1 - (1-mindist)*0.5;
	}
	printf("\n");


	int minboth = 99999;
	int maxboth;
	int factor = 200;
	int x1, x2, y1, y2, diffx, diffy;
	for (h = 0; h < numberOfLights; h++) {
		direction2xy(x1, y1, thelights[h].direction, size*factor);
		direction2xy(x2, y2, thelights[thelights[h].nearestNeighborNum].direction, size*factor);
		diffx = x1 - x2;
		if (diffx < 0) {diffx *= -1;}
		diffy = y1 - y2;
		if (diffy < 0) {diffy *= -1;}
		maxboth = diffx;
		if (diffy > diffx) {
			maxboth = diffy;
		}
		if (maxboth < minboth) {
			minboth = maxboth;
		}
	}
	int multiplier;
	if (minboth < factor) {
		multiplier = factor / minboth;
		multiplier *= 2;
	} else {
		multiplier = 1;
	}
	lightSize = size*multiplier;

	for (int k = 0; k < 255; k++) {
		keys[k] = false;
		keysdown = 0;
	}


	//build the voronoi diagram (discretely)
	int i, j;
	lightID = new int*[lightSize];
//	float ix, jy;
	for (i = 0; i < lightSize; i++) {
		lightID[i] = new int[lightSize];
	}
	float minDist, dot;
	Point3d unproj;
	for (i = 0; i < lightSize; i++) {
		printf("Bulding Voronoi %d/%d   %c", i, lightSize, i == lightSize-1 ? '\n' : '\r');
		for (j = 0; j < lightSize; j++) {
			if (!xy2direction(unproj, i, j, lightSize)) {
				lightID[j][i] = -1;
				continue;
			}
			minDist = -2;
			for (h = 0; h < numberOfLights; h++) {
				P3D_VDot(dot, thelights[h].direction, unproj);
				if (dot > minDist) {
					lightID[j][i] = h;
					minDist = dot;
				}
			}
		}
	}

	//build light neighboor graph
	horizons = new bool[numberOfLights];
	for (int hor = 0; hor < numberOfLights; hor++) {
		horizons[hor] = false;
	}
	graph = new Light_Graph*[numberOfLights];
	Light_Graph* runner;
	bool alreadyThere;
	int possibleNeighboor;

	for (h = 0; h < numberOfLights; h++) {
		graph[h] = new Light_Graph();
		graph[h]->firstChild = NULL;
		graph[h]->lightNumber = h;
		graph[h]->nextSibling = NULL;
	}
	int centerLight, d, e;
	multiplier = 1; //////////////////////////////////////////
	for (i = 0; i < lightSize; i++) {
		printf("building neighboor graph %d   %c", i, i == lightSize-1 ? '\n' : '\r');
		for (j = 0; j < lightSize; j++) { //for each point
			centerLight = lightID[j][i];
			if (centerLight == -1) {
				continue;
			}
			for (d = 1-(multiplier*2); d < (multiplier*2); d++) {
				if (i+d < 0 || i+d >= lightSize) {
					horizons[centerLight] = true;
					continue;
				}					
				for (e = 1-(multiplier*2); e < (multiplier*2); e++) {
					if (j+e < 0 || j+e >= lightSize) {
						horizons[centerLight] = true;
						continue;
					}
					possibleNeighboor = lightID[j+e][i+d]; //for each of it's 8 neighboors
					if (possibleNeighboor == -1) {
						horizons[centerLight] = true;
					}
					if (possibleNeighboor != centerLight && possibleNeighboor != -1) {
						alreadyThere = false;
						runner = graph[centerLight]->firstChild;
						while (runner != NULL) {			//if neighboor isn't already listed
							if (runner->lightNumber == possibleNeighboor) {
								alreadyThere = true;
								break;
							} else {
								runner = runner->nextSibling;
							}
						}
						if (!alreadyThere) {				//add to list
							runner = graph[centerLight]->firstChild;
							graph[centerLight]->firstChild = new Light_Graph();
							graph[centerLight]->firstChild->lightNumber = possibleNeighboor;
							graph[centerLight]->firstChild->firstChild = NULL;
							graph[centerLight]->firstChild->nextSibling = runner;
						}
					}
				}
			}
		}
	}

#ifndef RAYTRACER
	if (useGlut) {
		VVTrace::globalFlags[USE_TRUE] = true;
		theOneTrueTrace = new VVTrace();
		theOneTrueTrace->create(size, numberOfLights);
		theOneTrueTrace->trace();
		VVTrace::globalFlags[USE_TRUE] = false;

		int a;
		for (a = 0; a < 6; a++) {
			areas[a] = new Area(size);
		}

		for (a = 0; a < 6; a++) {
			activeArea = a;
			menu(101); //TrueTrace
		}
	}
#endif //RAYTRACER

}

#ifndef RAYTRACER
void Container::findGlutXY(int& glutx, int& gluty, int area, int areax, int areay) {
	glutx = ((area % 3) * size) + areax;
	gluty = ((area / 3) * size) + areay;
}
#endif //RAYTRACER

//looks up in the voronoi diagram to find the light at x,y scaled to lightSize
int Container::findLight(int first, int second) {
	int scale = lightSize / size;	
	return lightID[first*scale][second*scale];
}


//Area 0 is the top-left
int Container::findArea(int glutx, int gluty) {
	int incol = glutx / size;
	int inrow = gluty / size;
	return inrow*3 + incol;
}

#ifdef RAYTRACER
//can't use macros in here because types are not compatible
void Container::readData(char* canbenull) {
	numberOfLights = Frame.Number_Of_Light_Sources;
	size = Frame.Screen_Width;
	cameraLocation[0] = Frame.Camera->Location[0];
	cameraLocation[1] = Frame.Camera->Location[1];
	cameraLocation[2] = Frame.Camera->Location[2];

	cameraLookAt[0] = Frame.Camera->Look_At[0];
	cameraLookAt[1] = Frame.Camera->Look_At[1];
	cameraLookAt[2] = Frame.Camera->Look_At[2];

   LIGHT_SOURCE* thislight = Frame.Light_Sources;

	thelights = new Light[numberOfLights];
	for (int h = 0; h < numberOfLights; h++) {
		//color, direction
		thelights[h].color[0] = thislight->Colour[0];
		thelights[h].color[1] = thislight->Colour[1];
		thelights[h].color[2] = thislight->Colour[2];
		thelights[h].direction[0] = thislight->Center[0] / 1000.0;
		thelights[h].direction[1] = thislight->Center[1] / 1000.0;
		thelights[h].direction[2] = thislight->Center[2] / 1000.0;
		thelights[h].lightnum = h;
		thislight = thislight->Next_Light_Source;
	}
	Image_Info = new Depth_Normal*[size];
	for (int row = 0; row < size; row++) {
		Image_Info[row] = new Depth_Normal[size];
		for (int column = 0; column < size; column++) {
			Image_Info[row][column].depths = 0;
		}
	}

}
#endif //yes RAYTRACER


#ifndef RAYTRACER
void Container::readData(char* filebase) {
	char filename[FILE_NAME_LENGTH];
	
	printf("reading general file.\n");
	sprintf(filename, "%s_general3", filebase);
	FILE* general = fopen(filename, "r+");
	if (general == NULL) {
		printf("error: Could not open file: %s\n", filename);
		printf("Press a key to end program");
		getchar();
		exit(0);
	}
	fscanf(general, "numberOfLights %d\n", &numberOfLights);
	fscanf(general, "size %d\n", &size);
	fscanf(general, "cameraLocation %f %f %f\n", &cameraLocation[0],&cameraLocation[1],&cameraLocation[2] );
	fscanf(general, "cameraLookAt %f %f %f\n", &cameraLookAt[0], &cameraLookAt[1], &cameraLookAt[2]);
	int objnum;
	float r, g, b;
	while (fscanf(general, "object %d %f %f %f\n", &objnum, &r, &g, &b) == 4) {
		P3D_Make_Vector(objectColor[objnum], r, g, b);
	}
	fclose(general);

	float strongestSum = 0;
	float nextStrongest = 0;
	float thisSum = 0;
	Light_Graph*	sortSources = new Light_Graph();
	sortSources->lightNumber = 0;
	sortSources->nextSibling = NULL;
	Light_Graph* travelingSource;
	printf("reading light description file.\n");
	sprintf(filename, "%s_proximity3", filebase);
	FILE* proxim = fopen(filename, "r");
	thelights = new Light[numberOfLights];
	for (int h = 0; h < numberOfLights; h++) {
		//color, direction
		fscanf(proxim, "%f %f %f %f %f %f\n",
				&thelights[h].color[0], &thelights[h].color[1], &thelights[h].color[2],
				&thelights[h].direction[0], &thelights[h].direction[1], &thelights[h].direction[2]);
		thelights[h].lightnum = h;
		if (h != 0) {
			travelingSource = sortSources;
			while (travelingSource != NULL) {
				strongestSum = thelights[travelingSource->lightNumber].color[0]+
					thelights[travelingSource->lightNumber].color[1]+
					thelights[travelingSource->lightNumber].color[2];
				if (travelingSource->nextSibling != NULL) {
					nextStrongest = thelights[travelingSource->nextSibling->lightNumber].color[0]+
					thelights[travelingSource->nextSibling->lightNumber].color[1]+
					thelights[travelingSource->nextSibling->lightNumber].color[2];
				} else {
					nextStrongest = 0;
				}
				thisSum = thelights[h].color[0]+thelights[h].color[1]+thelights[h].color[2];
				if (thisSum > strongestSum && travelingSource == sortSources) {
					Light_Graph* insert = new Light_Graph();
					insert->lightNumber = h;
					insert->nextSibling = sortSources;
					sortSources = insert;
					break;
				}
				if (thisSum <= strongestSum && thisSum >= nextStrongest) {
					Light_Graph* insert = new Light_Graph();
					insert->lightNumber = h;
					insert->nextSibling = travelingSource->nextSibling;
					travelingSource->nextSibling = insert;
					break;
				}
				travelingSource = travelingSource->nextSibling;
			}
		}
	}
	fclose(proxim);

	int s = (numberOfLights / 100) + 1; //integer division is desirable
	if (s >= 20) {
		printf("Warning: not all strong sources will be traced\n");
		s = 20-1;
	}

	travelingSource = sortSources;
	for (int twenty = 0; twenty < 20; twenty++) {
		if (twenty < s) {
			strongestSources[twenty] = travelingSource->lightNumber;
			travelingSource = travelingSource->nextSibling;
		} else {
			strongestSources[twenty] = -1;
		}
	}

	sprintf(filename, "%s_data2", filebase);
	FILE* data = fopen(filename, "rb");
	Depth_Normal* ii;
	Image_Info = new Depth_Normal*[size];
	for (int row = 0; row < size; row++) {
		printf("reading row %d   %c", row, row == size-1 ? '\n' : '\r');
		Image_Info[row] = new Depth_Normal[size];
		for (int column = 0; column < size; column++) {
			ii = &Image_Info[row][column];
			fread(ii, 1, sizeof(Depth_Normal), data);
			if (ii->objectID) {
				ii->depths = new float[numberOfLights];
				fread(ii->depths, numberOfLights, sizeof(float), data);
			} else {
				ii->depths = 0;
			}
		}
	}
	printf("\n");
	fclose(data);
}


/*
If there is another pointer to this object, don't delete.
Just set it to NULL
*/
void Container::setTrace(int theArea, VVTrace* newValue) {
	if (newValue == theOneTrueTrace) {
		areas[theArea]->trace = newValue;
		return;
	}
	if (sameTrace(areas[theArea]->trace, newValue)) {
		delete newValue;
		return;
	}
	for (int c = 0; c < 6; c++) {
		if (sameTrace(areas[c]->trace, newValue)) {
			delete newValue;
			newValue = areas[c]->trace;
			break;
		}
	}
	int countt = 0;
	for (int a = 0; a < 6; a++) {
		if (areas[a]->trace == areas[theArea]->trace) {
			countt++;
		}
	}
	if (countt == 1) {
		if (areas[theArea]->trace != theOneTrueTrace) {
			delete areas[theArea]->trace;
		}
	} else {
		areas[theArea]->trace = NULL;
	}
	areas[theArea]->trace = newValue;
}

void Container::produceDiff(Area* intoArea, Area* area1, Area* area2) {
	float pigment;
	area1->unFocus();
	area2->unFocus();
	for (int row = 0; row < size; row++) {
		for (int col = 0; col < size; col++) {
			for (int c = 0; c < 3; c++) {
				pigment = area1->area[I(row,col)][c] - area2->area[I(row,col)][c];
				if (pigment < 0) {
					intoArea->area[I(row,col)][c] = -pigment;
				} else {
					intoArea->area[I(row,col)][c] = pigment;
				}
			}
		}
	}
}
#endif //RAYTRACER

#ifndef RAYTRACER

void Container::readInstructions(char* file) {
	FILE* fp = fopen(file, "r");

	if (fp != NULL) {
		printf("reading intructions:\n PRESS ENTER (focus this window) to execute each instuction\n");
	} else {
		printf("Could not open file %s\n", file);
		return;
	}
	char c;
	int entry, area, areax, areay, glutx, gluty, specialKey;
	c = '\n';
	while (c != EOF) {
		c = getc(fp);
		if (c == '/') {
			putchar(c);
			while (c != '\n') {
				c = getc(fp);
				putchar(c);
			}
		}
		if (c == 'p') {//passive
			fscanf(fp, "%d %d %d\n", &area, &areax, &areay);
			findGlutXY(glutx, gluty, area, areax, areay);
			printf("Passive motion to Area %d : %dx%d\n", area, areax, areay);
			getchar();
			passiveMotion(glutx, gluty);
		}
		if (c == 'r') { //right mouse button
			fscanf(fp, "%d\n", &entry);
			printf("Menu command %d\n", entry);
			getchar();
			menu(entry);
		}
		if (c == 'l') { //left mouse button
			fscanf(fp, "%d %d %d\n", &area, &areax, &areay);
			findGlutXY(glutx, gluty, area, areax, areay);
			if (leftDown) {
				printf("Left release at Area %d : %dx%d\n", area, areax, areay);
			} else {
				printf("Left press (and hold) at Area %d : %dx%d\n", area, areax, areay);
			}
			getchar();
			mouse(GLUT_LEFT, leftDown ? GLUT_UP : GLUT_DOWN, glutx, gluty);
		}
		if (c == EOF) {
			printf("Completed reading instructions.  Set focus to drawing window again\n");
			return;
		}
		if (c == 'k') { //for keyboard
			c = getc(fp);
			printf("keyboard press %c\n", c);
			getchar();
			keyboard(c, 0, 0);
		}
		if (c == 's') {
			fscanf(fp, "%d\n", &specialKey);
			printf("Press Special key %d\n", specialKey);
			getchar();
			special(specialKey, 0, 0);
		}
		display();
	}
	fclose(fp);
}

#endif //RAYTRACER

/////////////////////////////////////////////////////////////////////////
// GLUT functions
//////////////////////////////////////////////////////////////////////////

#ifndef RAYTRACER


void Container::display() {
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glMatrixMode(GL_MODELVIEW);
	double m[16] = {2.0/(double)glutWidth, 0, 0, 0,
					0, 2.0/(double)glutHeight, 0, 0,
					0, 0, 1, 0,
					-1, -1, 0, 1};
	glLoadIdentity();
	glMultMatrixd(m);
	glClearColor(0.3, 0.3, 0.3, 1);
	glClear(GL_COLOR_BUFFER_BIT);
	int area = 0;
	for (int rr = 0; rr < 2; rr++) {
		for (int cc = 0; cc < 3; cc++) {
			glRasterPos2i(cc*size, (1-rr)*size);
			glDrawPixels(size, size, GL_RGB, GL_FLOAT, areas[area++]->area);
		}
	}
	glutSwapBuffers();
}

// (0,0) for this (x,y) is top-left
void Container::passiveMotion(int x, int y) {
	if (record) {
		if (previousRecord[0] != 'p') {
			fprintf(recordFile, previousRecord);
		}
		sprintf(previousRecord, "p %d %d %d\n", findArea(x, y), x % size, y % size);
	}
	activeArea = findArea(x,y);
	int areax, areay;
	if (trackMouse && !areas[activeArea]->lockFocus) {
		areax = x % size;
		areay = y % size;
		areas[activeArea]->unFocus();
		areas[activeArea]->setFocus(areax, areay);
	}
	for (int a = 0; a < 6; a++) {
		if (passiveParent) {
			areas[a]->setParent(areas[activeArea]);
		} else if (areas[activeArea] == areas[a]->parentArea){
			if (trackMouse && !areas[activeArea]->lockFocus) {
				areas[a]->setParent(areas[a]->parentArea);
			}
		}
	}
	oldx = x;
	oldy = y;
	glutPostRedisplay();
}

void Container::reshape(int w, int h) {
	glutWidth = size*3;
	glutHeight = size*2;
	if (w != glutWidth || h != glutHeight) {
		glutReshapeWindow(glutWidth, glutHeight);
	}
	glViewport(0, 0, glutWidth, glutHeight);
	glutPostRedisplay();
}

void Container::motion(int x, int y) {
	oldx = x;
	oldy = y;
}

void Container::keyup(unsigned char key, int x, int y) {
	if (keys[key]) {
		keys[key] = false;
		keysdown--;
	}
}

void Container::keyboard(unsigned char key, int x, int y) {
	int f;
	if (!keys[key]) {
		keys[key] = true;
		keysdown++;
	}
	if (record) {
		fprintf(recordFile, previousRecord);
		sprintf(previousRecord, "k%c\n", key);
	}
	switch (key) {
	case '+':
		areas[activeArea]->brighten(2);
		break;
	case '-':
		areas[activeArea]->brighten(0.5);
		break;
	case 'l': //this is alower case L
		if (areas[activeArea]->lockFocus) {
			areas[activeArea]->lockFocus = false;
			printf("Area %d unlocked.\n", activeArea);
		} else {
			areas[activeArea]->lockFocus = true;
			printf("Area %d locked.\n", activeArea);
		}
		break;
	case 'v':
		areas[activeArea]->trace->print(stdout);
		break;
	case 'a':
		if (dragAllLights) {
			printf("Drag all lights: OFF\n");
			dragAllLights = false;
		} else {
			printf("Drag all lights: ON\n");
			dragAllLights = true;
		}
		break;
	case 'm':
		if (trackMouse) {
			printf("tracking mouse: OFF\n");
			trackMouse = false;
		} else {
			printf("tracking mouse: ON\n");
			trackMouse = true;
		}
		break;
	case 'p':
		if (passiveParent) {
			printf("Parent determined by mouse: OFF\n");
			passiveParent = false;
		} else {
			printf("Parent determined by mouse: ON\n");
			passiveParent = true;
		}
		break;
	case 't':
		areas[activeArea]->trace->tracePixel(areas[activeArea]->activey, areas[activeArea]->activex);
		break;
	case 'c':
		for (f = 0; f < NUMFLAGS; f++) {
			VVTrace::globalFlags[f] = areas[activeArea]->trace->flags[f];
		}
		break;
	case 'f':
		for (f = 0; f < NUMFLAGS; f++) {
			printf("%s=%c ", VVTrace::flagnames[f], VVTrace::globalFlags[f]? 'T' : 'F');
		}
		printf("\n");
		break;
///////////////////Flags//////////////////////
	case 'g':
	case '2':
		VVTrace::globalFlags[USE_GRID] = !VVTrace::globalFlags[USE_GRID];
		printf("globalFlags[USE_GRID] = %c\n", VVTrace::globalFlags[USE_GRID] ? 'T' : 'F');
		break;
	case 'b':
	case '4':
		VVTrace::globalFlags[USE_BFLOODING] = !VVTrace::globalFlags[USE_BFLOODING];
		printf("globalFlags[USE_BFLOODING] = %c\n", VVTrace::globalFlags[USE_BFLOODING] ? 'T' : 'F');
		break;
	case 'h':
	case '5':
		VVTrace::globalFlags[USE_HFLOODING] = !VVTrace::globalFlags[USE_HFLOODING];
		printf("globalFlags[USE_HFLOODING] = %c\n", VVTrace::globalFlags[USE_HFLOODING] ? 'T' : 'F');
		break;
	case 'u':
	case '6':
		VVTrace::globalFlags[USE_UFLOODING] = !VVTrace::globalFlags[USE_UFLOODING];
		printf("globalFlags[USE_UFLOODING] = %c\n", VVTrace::globalFlags[USE_UFLOODING] ? 'T' : 'F');
		break;
	case 'x':
	case '7':
		VVTrace::globalFlags[USE_HIGHC] = !VVTrace::globalFlags[USE_HIGHC];
		printf("globalFlags[USE_HIGHC] = %c\n", VVTrace::globalFlags[USE_HIGHC] ? 'T' : 'F');
		break;
	case 'w':
	case '8':
		VVTrace::globalFlags[USE_WARP] = !VVTrace::globalFlags[USE_WARP];
		printf("globalFlags[USE_WARP] = %c\n", VVTrace::globalFlags[USE_WARP] ? 'T' : 'F');
		break;
	case '3':
		VVTrace::globalFlags[USE_3OBJECT] = !VVTrace::globalFlags[USE_3OBJECT];
		printf("globalFlags[USE_3OBJECT] = %c\n", VVTrace::globalFlags[USE_3OBJECT] ? 'T' : 'F');
		break;
	case 'o':
	case '9':
		VVTrace::globalFlags[USE_OPTIMIZE] = !VVTrace::globalFlags[USE_OPTIMIZE];
		printf("globalFlags[USE_OPTIMIZE] = %c\n", VVTrace::globalFlags[USE_OPTIMIZE] ? 'T' : 'F');
		break;
	case 's':
	case '0':
		VVTrace::globalFlags[USE_STRONG] = !VVTrace::globalFlags[USE_STRONG];
		printf("globalFlags[USE_STRONG] = %c\n", VVTrace::globalFlags[USE_STRONG] ? 'T' : 'F');
		break;
	case '1':
		VVTrace::globalFlags[USE_ONEBACK] = !VVTrace::globalFlags[USE_ONEBACK];
		printf("globalFlags[USE_ONEBACK] = %c\n", VVTrace::globalFlags[USE_ONEBACK] ? 'T' : 'F');
		break;
	case 'r':
		VVTrace::globalFlags[USE_REAL] = !VVTrace::globalFlags[USE_REAL];
		printf("globalFlags[USE_REAL] = %c\n", VVTrace::globalFlags[USE_REAL] ? 'T' : 'F');
		break;
	}
	if (Container::useGlut) {
		glutPostRedisplay();
	} else {
		keyup(key, x, y);
	}
}

void Container::special(int key, int x, int y) {
	if (record) {
		fprintf(recordFile, previousRecord);
		sprintf(previousRecord, "s %d\n", key);
	}
	if (!trackMouse || areas[activeArea]->lockFocus) {
		x = areas[activeArea]->activex;
		y = areas[activeArea]->activey;
		switch (key) {
		case 100: //left
			x = areas[activeArea]->activex - 1;
			break;
		case 101: //up
			y = areas[activeArea]->activey - 1;
			break;
		case 102: //right
			x = areas[activeArea]->activex + 1;
			break;
		case 103: //down
			y = areas[activeArea]->activey + 1;
			break;
		}
		if (x < 0) {
			x = 0;
		}
		if (x > size-1) {
			x = size-1;
		}
		if (y < 0) {
			y = 0;
		}
		if (y > size-1) {
			y = size-1;
		}
		areas[activeArea]->unFocus();
		areas[activeArea]->setFocus(x, y);
		for (int a = 0; a < 6; a++) {
			if (passiveParent) {
				areas[a]->setParent(areas[activeArea]);
			} else if (areas[activeArea] == areas[a]->parentArea){
				areas[a]->setParent(areas[a]->parentArea);
			}
		}
	}
	glutPostRedisplay();
}

//produces a diff if both are !closeup
//creates an occluder if from !closeup to closeup

//drags a light occluder if both are closeup
//shows an occluder if from closeup to !closeup
void Container::dragFromTo(int fromAreaNum, int toAreaNum) {
	Area* fromArea = areas[fromAreaNum];
	Area* toArea = areas[toAreaNum];
	Point3d direction;
	DeepPixel *fromDP, *toDP;
	DeepPixel *fromParent, *toParent;

	if (fromArea->closeup) {
		fromParent = &(fromArea->parentArea->trace->deepBuffer[fromArea->parentArea->activey][fromArea->parentArea->activex]);
		if ((fromParent->numCorners > 0) && fromArea->corner != NULL) {
			fromDP = fromParent->corners[fromArea->corner-1];
		} else {
			fromDP = fromParent;
		}
	} else {
		fromDP = &(fromArea->trace->deepBuffer[fromArea->activey][fromArea->activex]);
	}

	if (toArea->closeup) {
		toParent = &(toArea->parentArea->trace->deepBuffer[toArea->parentArea->activey][toArea->parentArea->activex]);
		if ((toParent->numCorners > 0) && toArea->corner != NULL) {
			toDP = toParent->corners[toArea->corner-1];
		} else {
			toDP = toParent;
		}
	} else {
		toDP = &(toArea->trace->deepBuffer[toArea->activey][toArea->activex]);
	}

	//difference between two traces
	if (!fromArea->closeup && !toArea->closeup) { 
		produceDiff(toArea, fromArea, toArea);
		toArea->setType(DIFF);
		return;
	}

	//find out where in the lightspace a point of geometry lands
	if (!fromArea->closeup && toArea->closeup) {
		P3D_VSub(direction, fromDP->intersection, toDP->intersection);
		P3D_VNormalizeEq(direction);
		if (direction[1] <= 0) {
			printf("blocker warped below the horizon\n");
		} else {
			int x, y;
			direction2xy(x, y, direction, size);
			P3D_Make_Vector(toArea->area[I(y, x)], 1, 1, 1);
		}
	}

	//show where on the geometry the occluder of a light(s) are
	if (fromArea->closeup && !toArea->closeup) {
		Blocker_List* occluder = fromDP->occluders;
		while (occluder != NULL) {
			if (dragAllLights || (occluder->light == fromArea->activeLight)) {
				int x, y;
				project(x, y, occluder->occluder);
				P3D_Make_Vector(toArea->area[I(y, x)], 1, 1, 1);
			}
			occluder = occluder->next;
		}
	}

	//warp (or don't) blocker from the lightspace of one pixel to another
	Point3d toOccluder;
	Blocker_List* occluder;
	if (fromArea->closeup && toArea->closeup) {
		occluder = fromDP->occluders;
		while (occluder != NULL) {
			if (dragAllLights || (occluder->light == fromArea->activeLight)) {
				int x, y;
				if (toArea->parentArea->trace->flags[USE_WARP]) {
					P3D_VSub(toOccluder, occluder->occluder, toDP->intersection);
					P3D_VNormalizeEq(toOccluder);
				} else {
					P3D_Assign_Vector(toOccluder, thelights[occluder->light].direction);
				}
				direction2xy(x, y, toOccluder, size);
				for (int i = 0; i < 2; i++) {
					for (int j = 0; j < 2; j++) {
						if (y+j < size && x+i < size) {
							P3D_VScaleEq(toArea->area[I(y+j, x+i)], 2);
						}
					}
				}
			}
			occluder = occluder->next;
		}
	}
}


//a mouse click when trackMouse is off will place the text
//a mouse click when trackMouse is on will output the values of the clicked pixel
//a mouse drag will set a difference between two areas
void Container::mouse(int button, int state, int x, int y) {
	if (record) {
		fprintf(recordFile, previousRecord);
		sprintf(previousRecord, "l %d %d %d\n", findArea(x, y), x % size, y % size);
	}
	int theArea = findArea(x,y);
	if (button == GLUT_LEFT && state == GLUT_DOWN) {
		leftDown = true;
		if (!trackMouse) {
			areas[activeArea]->textx = x;
			areas[activeArea]->texty = (glutHeight-1) - y;
		}
		draggedArea = theArea;
		draggedx = x;
		draggedy = y;
	}
	if (button == GLUT_LEFT && state == GLUT_UP && leftDown) {
		leftDown = false;
		if (draggedArea != theArea) {
			dragFromTo(draggedArea, theArea);
		} else {
			if (areas[activeArea]->closeup) {
				int lnum = areas[activeArea]->activeLight;
				printf("area %d clicked with active light %d\n", activeArea, lnum);
				printf("R:%.3f  G:%.3f B:%.3f  (%.2f %.2f %.2f) [%.2f]\n",
					thelights[lnum].color[0], thelights[lnum].color[1], thelights[lnum].color[2],
					thelights[lnum].direction[0], thelights[lnum].direction[1], thelights[lnum].direction[2],
					thelights[lnum].farthestNeighborDist);
				if (areas[activeArea]->type == ALLLIGHTS) {
					int pcol, prow;
					direction2xy(pcol, prow, thelights[lnum].direction, areas[activeArea]->size);
					printf("Errors predicted invisible: %.0f, predicted visible: %.0f\n",
						areas[activeArea]->area[I(prow,pcol)][0]*numberOfLights,
						areas[activeArea]->area[I(prow,pcol)][2]*numberOfLights);
				}
			} else {
				printf("area %d clicked with active (%d, %d)\n", activeArea, areas[activeArea]->activex, areas[activeArea]->activey);
			}
		}
	}
	oldx = x;
	oldy = y;
	glutPostRedisplay();
}

void Container::menu(int entry) {
	int i;
	if (record) {
		fprintf(recordFile, previousRecord);
		if (entry != 12000) {
			sprintf(previousRecord, "r %d\n", entry);
		}
	}
	areas[activeArea]->unFocus();
	bool refill = true;
	if (entry < 100) {//regular stuff
		switch (entry) {
		case 1: 
			areas[activeArea]->normalize = true;
			break;
		case 2:
			areas[activeArea]->normalize = false;
			break;
		case 3:
			areas[activeArea]->brighten(2);
			refill = false;
			break;
		case 4:
			areas[activeArea]->brighten(0.5);
			refill = false;
			break;
		case 5:
			areas[activeArea]->setType(ONELIGHT);
			break;
		case 6:
			keyboard('c', areas[activeArea]->activex, areas[activeArea]->activey);
			break;
		case 7:
			areas[activeArea]->setType(ALLLIGHTS);
			break;
		}
		if (entry >= 10 && entry <= 20) {
			int theparent = entry % 10;
			areas[activeArea]->setParent(areas[theparent]);
		}
	}
	if (entry >= 100 && entry < 1000) { //not closeup
		switch (entry / 100) {
		case 1:
			areas[activeArea]->setType(TRACED);
			break;
		case 2:
			areas[activeArea]->setType(NUMVISIBLE);
			break;
		case 3:
			areas[activeArea]->setType(WORKDONE);
			break;
		case 4:
			areas[activeArea]->setType(ERRORS);
			break;
		}
		VVTrace* tr;
		switch (entry % 100) {
		case 1:
			tr = theOneTrueTrace;
			break;
		case 2: //2004 OPT
			for(i = 0; i < NUMFLAGS; i++) {
				VVTrace::globalFlags[i] = false;
			}
			VVTrace::globalFlags[USE_GRID] = true;
			VVTrace::globalFlags[USE_OPTIMIZE] = true;
			VVTrace::globalFlags[USE_BFLOODING] = true;
			VVTrace::globalFlags[USE_UFLOODING] = true;
			tr = new VVTrace();
			tr->create(size, numberOfLights);
			break;
		case 3:  //2004 GNU
			for(i = 0; i < NUMFLAGS; i++) {
				VVTrace::globalFlags[i] = false;
			}
			VVTrace::globalFlags[USE_GRID] = true;
			VVTrace::globalFlags[USE_UFLOODING] = true;
			tr = new VVTrace();
			tr->create(size, numberOfLights);
			break;
		case 4:
			tr = new VVTrace();
			tr->create(size, numberOfLights);
			break;
		case 5:
			VVTrace::globalFlags[USE_ALWAYS_VISIBLE] = true;
			tr = new VVTrace();
			tr->create(size, numberOfLights);
			VVTrace::globalFlags[USE_ALWAYS_VISIBLE] = false;
			break;
		}
		setTrace(activeArea, tr);
		if (!areas[activeArea]->trace->completed) {
			areas[activeArea]->trace->trace();
		}
	}
	if (entry >= 1000 && entry < 10000) { //closeup
		areas[activeArea]->corner = (entry / 1000) - 1;
		switch(entry % 100) {
		case 1:
			areas[activeArea]->setType(PREDICTION);
			break;
		}
	}
	if (entry >= 10000 && entry < 100000) { //readmenu
		char words[100];
		if (entry >= 12000) {//stop recording
			if (record) {
				fclose(recordFile);
				recordFile = NULL;
				record = false;
			} else {
				printf("Warning: Not recording to begin with\n");
			}
		} else if (entry >= 11000) {//start recording
			if (!record) {
				sprintf(words, "menu%d.txt", entry - 11000);
				sprintf(previousRecord, "p 0 0 0\n");
				recordFile = fopen(words, "w");
				record = true;
			} else {
				printf("Warning: already recording.  Previous recording will continue\n");
			}
		} else {
			sprintf(words, "menu%d.txt", entry - 10000);
			readInstructions(words);
		}
		refill = false;
	}
	if (refill) {
		areas[activeArea]->fillArea();
	}
}

void Container::project(int &x, int &y, Point3d location)
{
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	gluPerspective(53, 1, 0.5, 100);
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();
	gluLookAt(cameraLocation[0], cameraLocation[1], cameraLocation[2],
			cameraLookAt[0], cameraLookAt[1], cameraLookAt[2],
			0, 1, 0);
	double modelMatrix[16];
	double projMatrix[16];
	int viewport[4] = {0, 0, size, size};
	glGetDoublev(GL_MODELVIEW_MATRIX, modelMatrix);
	glGetDoublev(GL_PROJECTION_MATRIX, projMatrix);
	double winx, winy, winz;
	gluProject (location[0], location[1], location[2],
		modelMatrix, projMatrix, viewport, 
		&winx, &winy, &winz);
	x = size-winx;
	y = size-winy;
}
#endif //RAYTRACER




void Container::direction2xy(int& x, int& y, Point3d direction, int size) {
	Point3d p;
	projectStretch(p, direction);
	x = ((-1.0*p[0]+1.0)/2.0)*(size);
	y = ((p[1]+1.0)/2.0)*(size);
}

//keep
bool Container::xy2direction(Point3d& result, int x, int y, int size) {
	float ix, jy;
	ix = ((x+0.5)/(size*0.5) - 1)*-1;
	jy = ((y+0.5)/(size*0.5) - 1)*1;
	if (ix*ix + jy*jy > 1) {
		return false;
	}
	unprojectStretch(result, ix, jy);
	return true;
}
#define RAYTRACER2 //comment this for dominos
#ifdef RAYTRACER2
//keep
void Container::unprojectStretch(Point3d& result, double x, double y) {
	double z = sqrt(maxf(0.0, 1 - (x*x + y*y)));
	P3D_Make_Vector(result, x, z, y);
}


//keep
void Container::projectStretch(Point3d& result, Point3d direction) {
	double x, y;
	x = direction[0];
	y = direction[2];
	float dist = 1;
	P3D_Make_Vector(result, x, y, dist);
}

#else
void Container::unprojectStretch(Point3d& result, double x, double y) {
	double dist = sqrt(x*x + y*y);
	double ndist = sqrt(1 - (1-dist)*(1-dist));
	if (dist != 0) {
		x *= (ndist/dist);
		y *= (ndist/dist);
	}
	double z = sqrt(maxf(0.0, 1 - (x*x + y*y)));
	P3D_Make_Vector(result, x, z, y);
}

void Container::projectStretch(Point3d& result, Point3d direction) {
	double x, y;
	x = direction[0];
	y = direction[2];
	double dist = sqrt(x*x + y*y);
	double ndist;
	if (dist > 1) {
		dist = 1;
	}
	ndist =  1 - sqrt(1-dist*dist);
	x = x * (ndist/dist);
	y = y * (ndist/dist);
	P3D_Make_Vector(result, x, y, dist);
}
#endif //if else RAYTRACER

//keep
float Container::maxf(float a, float b) {
	if (a > b) {
		return a;
	} else {
		return b;
	}
}

//keep
int Container::intlog2(int num) {
	switch(num) {
	case 1024:
		return 10;
	case 512:
		return 9;
	case 256:
		return 8;
	case 128:
		return 7;
	case 64:
		return 6;
	case 32:
		return 5;
	case 16:
		return 4;
	case 8:
		return 3;
	case 4:
		return 2;
	case 2:
		return 1;
	case 1:
		return 0;
	}
	return -1;
}

#ifndef RAYTRACER
//keep
bool Container::sameTrace(VVTrace* one, VVTrace* other) {
	if (one == NULL) {
		return false;
	}
	if (other == NULL) {
		return false;
	}
	bool sofar = true;
	if (one->flags[USE_TRUE] && other->flags[USE_TRUE]) {
		return true;
	}
	if (one->flags[USE_ALWAYS_VISIBLE] && other->flags[USE_ALWAYS_VISIBLE]) {
		return true;
	}
	for (int flag = 0; flag < NUMFLAGS; flag++) {
		if (one->flags[flag] != other->flags[flag]) {
			sofar = false;
		}
	}
	return sofar;
}
#endif //RAYTRACER

void Container::rotatelights(float degrees)
{
	float rads = degrees * 3.14159265359 / 180;
	for (int h = 0; h < numberOfLights; h++) {
		float x = thelights[h].direction[0];
		float z = thelights[h].direction[2];
		thelights[h].direction[0] = x * cos(rads) - z * sin(rads);
		thelights[h].direction[2] = x * sin(rads) + z * cos(rads);
	}
}
