// Aner Ben-Artzi - Efficient Coherence-Based Visibility Sampling (CBVS) of Environment Maps
// Columbia University 2004
//
// VVTrace.h: interface for the VVTrace class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_TRACE_H__C8BB4819_452E_4745_9E3A_D515C17ECB68__INCLUDED_)
#define AFX_TRACE_H__C8BB4819_452E_4745_9E3A_D515C17ECB68__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "Container.h"


enum OBJECTCHECK {
	ID,
	DIST,
	NOOBJECT,
	NUMOBJECTCHECK
};

#ifndef RAYTRACER

class CompareTrace;

enum DEPTHTYPE {
	TRUE,
	NEAREST,
	ACTUAL,
	NUMDEPTHTYPE
};
#endif //RAYTRACER

#define STEPSIZE 16
#define STAGES 5

enum flags_enum {
	USE_ONEBACK,		//makes scanline use only one previous pixel (only valid with !USE_GRID)1 1
	USE_GRID,			//chooses between coarse-to-fine or scanline order						g 2
	USE_3OBJECT,		//allows only 3 to fall on same object									3 3
	USE_BFLOODING,		//boundary flooding (not includeing horizon)							b 4
	USE_HFLOODING,		//horizon flooding														h 5
	USE_UFLOODING,		//uncertianty flooding (not valid with USE_ONEBACK)						u 6
	USE_WARP,			//warps the blockers to local pixel (!warp means check curvature)		w 8
	USE_HIGHC,			//additional check for high curvature									x 7
	USE_OPTIMIZE,		//sets !warp, on cstep==2												o 9
	USE_STRONG,			//forces tracing of 1%+1 strongest sources								s 0			
	USE_TRUE,			//overrides everything - always traces rays								 
	USE_ALWAYS_VISIBLE,	//overrides everything - always predicts light as visible				 
	USE_REAL,			//after making a prediction, sets the correct visibility values			r
	NUMFLAGS
};


class VVTrace  
{
public:
	static char* flagnames[NUMFLAGS];

friend class CompareTrace;


public:
	void fastRender();
	void traceStrongestSources();
	void retraceTraced();
	void floodHorizon();
	void traceUnpredictable();
	bool notEnoughCorners(int numcorners);
	void tracePrimaryRay();
	bool flags[NUMFLAGS];
	static bool globalFlags[NUMFLAGS];
	void addOccluder(int row, int col, int light, Point3d& occluder);
	void removeOccluder(int row, int col, int light);
	void writeDiffPPM(char *filename, VVTrace* other);
	void writePPM(char* filename);
	void floodOutwards(Light_Graph* mightBeAffected, Light_Graph* affectedTail);
	bool highCurvature(int row, int col);
	bool sameObject();
	void floodGray();
	void floodBoundry();
	void floodIntersection();

	void print(FILE* fp);
	void subprint(FILE* fp, int num, float under, char* text);
	int* cvisible;
	int* cblocked;
	float* bestfall;
	int* usedNeighbors;
	int usedNeighborsPointer;
	Blocker_List** bestoccluder;
	void setTitle();
#ifndef RAYTRACER
	void fillExtraInfo();
	int* predviserrors;
	int* predinviserrors;
#endif //RAYTRACER

 	void findScanCorners(int row, int col);
 	void findGridCorners(int row, int col);
	char title[200];
	void tracePixel(int row, int col);
	void trace();
	Point3d** traced;
	virtual void create(int thesize, int thenumlights);
	VVTrace();
	virtual ~VVTrace();
	bool* visible;
	DeepPixel** deepBuffer;
#ifndef RAYTRACER
	BRDF brdf;
	static void doBRDF(Point3d& result, BRDF thebrdf, int thelight, Point3d& theEyeRay, Point3d& theNormal, int objectID);
#else
	static void doBRDF(Point3d& result, int thelight, Point3d& theNormal, int objectID);
#endif //RAYTRACER

	bool completed;

	int numTraced;
	int numAboveHorizon;
	int numVisIncor;
	int numInvisIncor;
	int numVisCor;
	int numInvisCor;
	int numTracedVis;
	int numTracedInvis;
	int numBHorizon;
	int numPixelsHit;
	int numFloodChange;
	int numFloodSame;
	int numNotEnoughNeighbors;
	char* className;
	int numlights;


protected:
	bool traceShadowRay(Point3d& result);
	bool traceShadowRay();
	bool traceShadowRay(Point3d& result, int row, int col, int light);
	bool traceShadowRay(int row, int col, int light);
	void predictAllLights();
	void predictAllLightsNoWarp();
	void traceAlternatingGrid();
	void traceScanline();
	int size;
	void addSingleBlockage(Blocker_List* possibleOccluder);
	int ccorner;
	int clight;
	int crow;
	int ccol;
	int cstep;
	void tracePixel();
};

#endif // !defined(AFX_TRACE_H__C8BB4819_452E_4745_9E3A_D515C17ECB68__INCLUDED_)
