// Aner Ben-Artzi - Efficient Coherence-Based Visibility Sampling (CBVS) of Environment Maps
// Columbia University 2004
//

//Command line description:
//VVTrace.h defines the flags and their associated keys
//to run a rendering with grid=true, warp=true, uncertainty=true, type:
//c:/visview [data file basename] gwu

#include "Container.h"
#include "VVTrace.h"

//keep
int main(int argc, char* argv[]) {

	bool output;

	char filenames[50][FILE_NAME_LENGTH];
	int i = 0;
	int total = 0;
	if (argc == 1) {
		std::vector<char[FILE_NAME_LENGTH]> options;
		FILE* scenes = fopen("scenes.txt", "r");
		if (scenes != NULL) {
			while (fscanf(scenes, "%s\n", filenames[total]) == 1) {
				total++;
			}
			fclose(scenes);
			for (i = 0; i < total; i++) {
				printf("(%d) %s\n", i, filenames[i]);
			}
		}
		printf("(%d) Other\n", i);
		printf("Choose a scene: ");
		int chosen;
		scanf("%d", &chosen);
		if (chosen == i) {
//			printf("Type the full path up to .bmp: ");
//			scanf("%s", argv1);
//			Container::readData(argv1);
		} else {
			Container::readData(filenames[chosen]);
		}
		output = false;
	} else {
		Container::useGlut = false;
		Container::readData(argv[1]);
		output = true;
	}
	Container::init();

	if (Container::useGlut) {
		argc = 1;
		glutInit( &argc, argv );
		glutInitWindowPosition( 100, 100 );
		glutInitWindowSize(Container::glutWidth, Container::glutHeight);
		glutInitDisplayMode( GLUT_RGB | GLUT_DOUBLE);

		glutCreateWindow( "Visview for Coherence-Based Visibility Sampling" );
		glutDisplayFunc( Container::display );
		glutKeyboardFunc( Container::keyboard );
		glutKeyboardUpFunc(Container::keyup);

		glutSpecialFunc( Container::special );
		glutMouseFunc( Container::mouse );
		glutReshapeFunc( Container::reshape );
		glutPassiveMotionFunc( Container::passiveMotion );
		glutMotionFunc(Container::motion);

		int normalize = glutCreateMenu(Container::menu);
		glutAddMenuEntry("Normalize", 1);
		glutAddMenuEntry("Un-Normalize", 2);

		int brightness = glutCreateMenu(Container::menu);
		glutAddMenuEntry("Brighten", 3);
		glutAddMenuEntry("Darken", 4);


		//to add a new type of trace, increment the constant by 1 in both places
		//then add it as a submenu after all the others below.
		//also add it to Container::menu after the others.
		int traceMenus[5];
		for (int t = 0; t < 5; t++) {
			traceMenus[t] = glutCreateMenu(Container::menu);
			glutAddMenuEntry("Trace", 101+t);
			glutAddMenuEntry("Num Visible Lights", 201+t);
			glutAddMenuEntry("Work done", 301+t);
			glutAddMenuEntry("Errors", 401+t);
		}

		int submens[5];
		for (int submen = 0; submen < 5; submen++) {
			submens[submen] = glutCreateMenu(Container::menu);
			glutAddMenuEntry("predictions", submen*1000 + 1001);
		}
		int parentmenu = glutCreateMenu(Container::menu);
			glutAddMenuEntry("Area 0", 10); 
			glutAddMenuEntry("Area 1", 11); 
			glutAddMenuEntry("Area 2", 12); 
			glutAddMenuEntry("Area 3", 13); 
			glutAddMenuEntry("Area 4", 14); 
			glutAddMenuEntry("Area 5", 15);
		
		int readMenu = glutCreateMenu(Container::menu);
		char words[100];
		int readmenu;
		for (readmenu = 0; readmenu < 10; readmenu++) {
			sprintf(words, "read \"menu%d.txt\"", readmenu);
			glutAddMenuEntry(words, 10000 + readmenu);
		}
		int startRecord = glutCreateMenu(Container::menu);
		for (readmenu = 0; readmenu < 10; readmenu++) {
			sprintf(words, "record to \"menu%d.txt\"", readmenu);
			glutAddMenuEntry(words, 11000 + readmenu);
		}

		int recordMenu = glutCreateMenu(Container::menu);
		glutAddMenuEntry("Stop recording", 12000);
		glutAddSubMenu("Start recording>", startRecord);
		glutAddSubMenu("Play recording>", readMenu);


		glutCreateMenu(Container::menu);
		glutAddSubMenu("Record Macro", recordMenu);
		glutAddSubMenu("Normalize", normalize);
		glutAddSubMenu("Brightness", brightness);
		glutAddMenuEntry("One Light", 5);
		glutAddMenuEntry("Capture the Flags", 6);
		glutAddMenuEntry("All Lights Errors", 7);
		glutAddSubMenu("Set Parent to:", parentmenu);
		glutAddSubMenu("True Trace", traceMenus[0]);
		glutAddSubMenu("CBVS 2004 OPT", traceMenus[1]);
		glutAddSubMenu("CBVS 2004 GNU", traceMenus[2]);
		glutAddSubMenu("Trace current flags", traceMenus[3]);
		glutAddSubMenu("Trace Visible", traceMenus[4]);
		glutAddSubMenu("Center", submens[0]);
		glutAddSubMenu("topLEFT corner (0)", submens[1]); //corner1
		glutAddSubMenu("TOPright corner (1)", submens[2]); //corner2
		glutAddSubMenu("BOTTOMleft corner (2)", submens[3]); //corner3
		glutAddSubMenu("bottomRIGHT corner (3)", submens[4]); //corner4
		glutAttachMenu(GLUT_RIGHT_BUTTON);
	} //end of if useGlut

	VVTrace* test;

	char file[FILE_NAME_LENGTH];
	if (output) {
		FILE* textfp = fopen("run.txt", "a");
		fprintf(textfp, "\n%s\n", argv[1]);
		int i = 0;
		if (argc == 3) {
			while (argv[2][i] != NULL) {
				Container::keyboard(argv[2][i], 0, 0);
				i++;
			}
		} else {
			VVTrace::globalFlags[USE_TRUE] = true;
		}
		test = new VVTrace();
		test->create(Container::size, Container::numberOfLights);
		test->trace();
		sprintf(file, "%s.ppm", argv[1]);
		test->writePPM(file);
		test->print(textfp);
		delete test;
		fclose(textfp);
	} else {
		glutMainLoop();
	}
	return 0;
}
