// Aner Ben-Artzi - Efficient Coherence-Based Visibility Sampling (CBVS) of Environment Maps
// Columbia University 2004
// mostly cut-and-paste from POV-Ray

#ifndef P3D_VECTOR_H
#define P3D_VECTOR_H

enum
{
	P3D_X = 0,
	P3D_Y = 1,
	P3D_Z = 2,
	P3D_T = 3
};

/* Misc. Vector Math Macro Definitions */
#define P3D_Assign_Vector(d,s)  memcpy((d),(s),sizeof(Point3d))
#define P3D_Make_Vector(v,a,b,c) { (v)[P3D_X]=(a);(v)[P3D_Y]=(b);(v)[P3D_Z]=(c); }
#define P3D_Make_Gray_Vector(v, g) { (v)[P3D_X]=(g);(v)[P3D_Y]=(g);(v)[P3D_Z]=(g); }

/* Vector Add */
#define P3D_VAdd(a, b, c) {(a)[P3D_X]=(b)[P3D_X]+(c)[P3D_X];(a)[P3D_Y]=(b)[P3D_Y]+(c)[P3D_Y];(a)[P3D_Z]=(b)[P3D_Z]+(c)[P3D_Z];}
#define P3D_VAddEq(a, b) {(a)[P3D_X]+=(b)[P3D_X];(a)[P3D_Y]+=(b)[P3D_Y];(a)[P3D_Z]+=(b)[P3D_Z];}

/* Vector Subtract */
#define P3D_VSub(a, b, c) {(a)[P3D_X]=(b)[P3D_X]-(c)[P3D_X];(a)[P3D_Y]=(b)[P3D_Y]-(c)[P3D_Y];(a)[P3D_Z]=(b)[P3D_Z]-(c)[P3D_Z];}
#define P3D_VSubEq(a, b) {(a)[P3D_X]-=(b)[P3D_X];(a)[P3D_Y]-=(b)[P3D_Y];(a)[P3D_Z]-=(b)[P3D_Z];}

/* Scale - Multiply Vector by a Scalar */
#define P3D_VScale(a, b, k) {(a)[P3D_X]=(b)[P3D_X]*(k);(a)[P3D_Y]=(b)[P3D_Y]*(k);(a)[P3D_Z]=(b)[P3D_Z]*(k);}
#define P3D_VScaleEq(a, k) {(a)[P3D_X]*=(k);(a)[P3D_Y]*=(k);(a)[P3D_Z]*=(k);}

/* Dot Product - Gives Scalar angle (a) between two vectors (b) and (c) */
#define P3D_VDot(a, b, c) {a=(b)[P3D_X]*(c)[P3D_X]+(b)[P3D_Y]*(c)[P3D_Y]+(b)[P3D_Z]*(c)[P3D_Z];}

/* Square a Vector (b) and Assign to another Vector (a) */
#define P3D_VSquareTerms(a, b) {(a)[P3D_X]=(b)[P3D_X]*(b)[P3D_X];(a)[P3D_Y]=(b)[P3D_Y]*(b)[P3D_Y];(a)[P3D_Z]=(b)[P3D_Z]*(b)[P3D_Z];}

/* Vector Length - returs Scalar Euclidean Length (a) of Vector (b) */
#define P3D_VLength(a, b) {a=sqrt((b)[P3D_X]*(b)[P3D_X]+(b)[P3D_Y]*(b)[P3D_Y]+(b)[P3D_Z]*(b)[P3D_Z]);}

/* Vector Distance - returs Scalar Euclidean Distance (a) between two
 * points/Vectors (b) and (c) */
#define P3D_VDist(a, b, c) {Point3d _tmp; P3D_VSub(_tmp, b, c); P3D_VLength(a, _tmp);}

#define P3D_EPSILON 0.000001

/* Normalize a Vector - returns a vector (length of 1) that points at (b) */
#define P3D_VNormalize(a,b) {float _tmp=1/sqrt((b)[P3D_X]*(b)[P3D_X]+(b)[P3D_Y]*(b)[P3D_Y]+(b)[P3D_Z]*(b)[P3D_Z]);(a)[P3D_X]=(b)[P3D_X]*_tmp;(a)[P3D_Y]=(b)[P3D_Y]*_tmp;(a)[P3D_Z]=(b)[P3D_Z]*_tmp;}
#define P3D_VNormalizeEq(a) {float _tmp=1/sqrt((a)[P3D_X]*(a)[P3D_X]+(a)[P3D_Y]*(a)[P3D_Y]+(a)[P3D_Z]*(a)[P3D_Z]);(a)[P3D_X]*=_tmp;(a)[P3D_Y]*=_tmp;(a)[P3D_Z]*=_tmp;}

/* Calculate the sum of the sqares of the components of a vector.  (the square of its length) */
#define P3D_VSumSqr(a)  ((a)[P3D_X]*(a)[P3D_X] + (a)[P3D_Y]*(a)[P3D_Y] + (a)[P3D_Z]*(a)[P3D_Z])

#endif


