// Aner Ben-Artzi - Efficient Coherence-Based Visibility Sampling (CBVS) of Environment Maps
// Columbia University 2004
//


#ifndef VV_HEADER_H
#define VV_HEADER_H

#include <cmath>
#include <stdio.h>

#ifdef RAYTRACER
#include "C:/projects/pov/povwin-3.5/source/frame.h"
#include "C:/projects/pov/povwin-3.5/source/vector.h"
#include "C:/projects/pov/povwin-3.5/source/RAYTRACER.h"
#endif

#ifndef RAYTRACER
#include <stdlib.h>
#include <stdarg.h>
#include <memory.h>
#include <vector>
#include "gl/glut.h"
#endif //RAYTRACER

#include "p3d_vector.h"
typedef float Point3d [3];
#define LOWER_HEMISPHERE
#ifndef RAYTRACER

enum depths {
	BELOW_HORIZON = -2,
	VISIBLE = -1
};


enum brdf {
	DIFFUSE,
	PHONG,
	BINARY, //no ndotl
	EXPANDED, //dim lights get dimmer
	MONOCHROME //no ndotl or light color (assumed to be (1, 1, 1)
};
enum areatype {
	TRACED,
	NUMVISIBLE,
	WORKDONE,
	ERRORS,
	DIFF,
	ONELIGHT,
	PREDICTION,
	ALLLIGHTS
};
typedef enum brdf BRDF;
typedef enum areatype AreaType;

#endif //RAYTRACER

class Light
{
public:
	Point3d color;
	Point3d direction;
	int lightnum;
	float farthestNeighborDist;
	int nearestNeighborNum;
	float nearestNeighborDist;
};

class Light_Graph
{
public:
	int lightNumber;
	Light_Graph* nextSibling;
	Light_Graph* firstChild;
};

enum predictions {
	TOBETRACED,
	WASTRACEDVIS,
	WASTRACEDINVIS,
	BHORIZON,
	PREDVIS,
	PREDINVIS
};

#ifndef RAYTRACER
enum extrainfos {
	CALCULATED,
	CORRECT,
	INCORRECT,
	REALDATA
};
#endif //RAYTRACER


class Blocker_List
{
public:
	Point3d occluder;
	int light;
	Blocker_List* next;
};

class DeepPixel
{
public:
	int objectID;
	Point3d intersection;
	Point3d normal;
	DeepPixel* corners[4];
	int numCorners;
	char* prediction;
	char* extrainfo;
	Blocker_List* occluders;
	int row;
	int col;
};

class Depth_Normal
{
public:
	int objectID;
	Point3d true_color;
	Point3d iPoint;
	Point3d iNormal;
	float* depths;
};


class Trace;

#ifndef RAYTRACER
#define FILE_NAME_LENGTH 1024
#define PI 3.1415926535897932384626433832795
#define FLOODSPEED 5

#endif //RAYTRACER

#endif